/* insmod.c: insert a module into the kernel.
    Copyright (C) 2001  Rusty Russell.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <asm/unistd.h>

static void print_usage(const char *progname)
{
	fprintf(stderr,
		"Usage: %s [--name=<modulename>] filename [args]\n",
		progname);
	exit(1);
}

/* We use error numbers in a loose translation... */
static const char *moderror(int err)
{
	switch (err) {
	case ENOEXEC:
		return "Invalid module format";
	case ENOENT:
		return "Unknown symbol in module";
	default:
		return strerror(err);
	}
}

int main(int argc, char *argv[])
{
	unsigned int i;
	int fd;
	long int ret;
	struct stat st;
	unsigned long len;
	void *map;
	char *name = NULL, *filename, *options = strdup("");

	/* Don't use getopt: keep this tiny */
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "--") == 0) break;
		if (strncmp(argv[i], "--name=", 7) == 0) {
			name = argv[i] + 7;
			memcpy(argv + i, argv + i + 1,
			       (argc - i) * sizeof(char *));
			argc--;
			break;
		}
		/* No other options known. */
		if (strncmp(argv[i], "--", 2) == 0)
			print_usage(argv[0]);
	}
	filename = argv[1];
	if (!filename)
		print_usage(argv[0]);

	if (!name) {
		char *ext;
		/* Strip path and .o off filename to give name */
		name = strrchr(filename, '/');
		if (!name) name = filename;
		else name++;
		name = strdup(name);
		ext = strrchr(name, '.');
		if (ext) *ext = '\0';
	}

	/* Rest is options */
	for (i = 2; i < argc; i++) {
		options = realloc(options,
				  strlen(options) + 2 + strlen(argv[i]) + 2);
		/* Spaces handled by "" pairs, but no way of escaping
                   quotes */
		if (strchr(argv[i], ' '))
			strcat(options, "\"");
		strcat(options, argv[i]);
		if (strchr(argv[i], ' '))
			strcat(options, "\"");
		strcat(options, " ");
	}

	fd = open(filename, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "Can't open `%s': %s\n",
			filename, strerror(errno));
		exit(1);
	}

	fstat(fd, &st);
	len = st.st_size;
	map = mmap(NULL, len, PROT_READ, MAP_SHARED, fd, 0);
	if (map == MAP_FAILED) {
		fprintf(stderr, "Can't map `%s': %s\n",
			filename, strerror(errno));
		exit(1);
	}

	ret = syscall(__NR_init_module, name, map, len, options);
	if (ret != 0) {
		fprintf(stderr, "Error inserting `%s': %li %s\n",
			filename, ret, moderror(errno));
		exit(1);
	}
	exit(0);
}
